@@ -7,13 +7,37 @@ import contextlib
 import enum
 import typing as t
 from dataclasses import dataclass
-from termcolor import colored
+
+try:
+  from termcolor import colored
+except ImportError:
+  def colored(s, *a, **kw) -> str: return str(s)  # type: ignore
 
 from nr.parsing.core import rules
 from nr.parsing.core.scanner import Cursor
 from nr.parsing.core.tokenizer import ProxyToken as _ProxyToken, RuleSet, Tokenizer
 
 
+@dataclass
+class Grammar:
+  """
+  Grammar settings.
+  """
+
+  #: Whether unparenthesized function calls are allowed.
+  unparen_calls: bool = True
+
+  #: Whether keyword arguments may be specified using colons (`:`) as well as equal signs (`=`).
+  colon_kwargs: bool = True
+
+  #: Accept function arguments with comma separation.
+  nocomma_args: bool = True
+
+  #: Whether the `def varname = ...` syntax is allowed and understood. This is an important
+  #: syntax feature when enabling the #NameRewriter with #TranspileOptions.closure_target.
+  local_def: bool = False
+
+
 class Token(enum.Enum):
   Eof = enum.auto()
   Indent = enum.auto()
@@ -59,19 +83,21 @@ class ParseMode(enum.IntFlag):
   DEFAULT = 0
 
   #: The currently parsed expression is grouped in parenthesis and may wrap over lines.
-  GROUPED = 1
+  GROUPED = 1 << 0
 
   #: The currently parsed expression is the outter parenthesis of a function call.
-  FUNCTION_CALL = (1 << 1)
+  FUNCTION_CALL = 1 << 1
 
   #: The currently parsed expression is an argument to a function call.
-  CALL_ARGS = (1 << 2)
+  CALL_ARGS = 1 << 2
 
 
 @dataclass
 class SyntaxError(Exception):
   """
   Specialized Craftr DSL syntax error (the internal SyntaxError class is weird).
+
+  If the `termcolor` module is installed, the error message will be color coded.
   """
 
   message: str
@@ -122,7 +148,11 @@ class Closure:
 
 @dataclass
 class RewriteResult:
-  #: The rewritten Python code.
+  """
+  The result of rewriting Craftr DSL code to pure Python code.
+  """
+
+  #: The Craftr DSL code rewritten as Python code.
   code: str
 
   #: The closures extracted from the code.
@@ -130,29 +160,32 @@ class RewriteResult:
 
 
 class Rewriter:
+  """
+  This class rewrites Craftr DSL code to pure Python code. Closures are extracted from the code
+  and replaced with whitespace where appropriate to keep line and column numbers of the code in
+  tact as much as possible (not always fully accurate).
+  """
 
   UNARY_OPERATORS = frozenset(['not', '~'])
   BINARY_OPERATORS = frozenset(['-', '+', '*', '**', '/', '//', '^', '|', '&', '.', '==', '<=', '>=', '<', '>', 'is', '%'])
   PYTHON_BLOCK_KEYWORDS = frozenset(['class', 'def', 'if', 'elif', 'else', 'for', 'while', 'with'])
 
-  def __init__(self, text: str, filename: str, supports_local_def: bool) -> None:
+  def __init__(self, text: str, filename: str, grammar: t.Optional[Grammar] = None) -> None:
     """
     # Arguments
     text: The Craftr DSL code to parse and turn into an AST-like structure.
     filename: The filename where the DSL code is from.
-    supports_local_def: Whether the `def varname = ...` syntax is allowed and understood. This is an important
-      syntax feature when enabling the #NameRewriter with #TranspileOptions.closure_target.
     """
 
     self.tokenizer = Tokenizer(rule_set, text)
     self.filename = filename
-    self.supports_local_def = supports_local_def
+    self.grammar = grammar or Grammar()
     self._closure_stack: t.List[str] = []  #: Used to construct nested closure names.
     self._closure_counter = 0  #: Used to assign a unique number to every closure.
     self._closures: t.Dict[str, Closure] = {}
 
   @contextlib.contextmanager
-  def _playing_games(self) -> t.Iterator[t.Callable[[], None]]:
+  def _lookahead(self) -> t.Iterator[t.Callable[[], None]]:
     """
     Context manager to save the current tokenizer and closure state and restore it on exit. This is
     useful for lookaheads, like :meth:`_test_dict`. If the returned callable is called, the
@@ -373,19 +406,33 @@ class Rewriter:
 
     token = ProxyToken(self.tokenizer)
     code = ''
+    upsert_comma = False
     while True:
       code += self._consume_whitespace(mode)
-      code += self._rewrite_expr(mode=mode)
+      with self._lookahead() as commit:
+        try:
+          code += self._rewrite_expr(mode=mode)
+        except SyntaxError:
+          break
+        commit()
       code += self._consume_whitespace(mode)
-      if mode & ParseMode.CALL_ARGS and token.is_control('='):
+      if mode & ParseMode.CALL_ARGS and (token.is_control('=') or (self.grammar.colon_kwargs and token.is_control(':'))):
         code += '='
         token.next()
         # TODO(NiklasRosenstein): This may be problematic in unparenthesised calls?
         code += self._rewrite_expr(mode=mode)
-      if not token.is_control(','):
+      if token.is_control(','):
+        code += ','
+        token.next()
+        upsert_comma = False
+      elif mode & ParseMode.CALL_ARGS and self.grammar.nocomma_args:
+        code += ','
+        upsert_comma = True
+      else:
         break
-      code += ','
-      token.next()
+
+    if upsert_comma:
+      code = code.rstrip(',')
     return code
 
   def _rewrite_atom(self, mode: ParseMode = ParseMode.DEFAULT) -> str:
@@ -450,7 +497,7 @@ class Rewriter:
     token = ProxyToken(self.tokenizer)
     assert token.is_control('{'), False
 
-    with self._playing_games():
+    with self._lookahead():
       token.next()
       self._consume_whitespace(True, False)
       try:
@@ -528,7 +575,7 @@ class Rewriter:
       token.next()
       code += '=' + self._consume_whitespace(newlines=False) + self._rewrite_items(ParseMode.DEFAULT)
 
-    elif token and not token.is_ignorable(True) and not token.is_control(')]}:'):  # Unparenthesises functionc all
+    elif token and not token.is_ignorable(True) and not token.is_control(')]}:') and self.grammar.unparen_calls:
       if code[-1].isspace():
         code = code[:-1]
       # TODO(NiklasRosenstein): We may want to indicate here that we're parsing call arguments,
@@ -546,7 +593,7 @@ class Rewriter:
     token = ProxyToken(self.tokenizer)
     assert token.tv == (Token.Name, 'def'), token
 
-    with self._playing_games() as commit:
+    with self._lookahead() as commit:
       token.next()
       self._consume_whitespace(False)
       if token.type != Token.Name:
@@ -581,7 +628,7 @@ class Rewriter:
     code += token.value
     token.next()
 
-    if self.supports_local_def and token.tv == (Token.Name, 'def') and (defcode := self._test_local_def()):
+    if self.grammar.local_def and token.tv == (Token.Name, 'def') and (defcode := self._test_local_def()):
       return code + defcode
 
     if token.type == Token.Name and token.value in self.PYTHON_BLOCK_KEYWORDS:
@@ -619,4 +666,9 @@ class Rewriter:
     return code
 
   def rewrite(self) -> RewriteResult:
+    """
+    Rewrite the code and return the #RewriteResult. This can be interpreted by the
+    #craftr.dsl.transpiler.ClosureRewriter to re-inject the code for closures.
+    """
+
     return RewriteResult(self._rewrite_stmt_block(), self._closures)
